@@ -8,6 +8,7 @@ import portage
 from portage import _encodings
 from portage import _unicode_encode
 from portage import os
+from portage.const import BASH_BINARY
 import fcntl
 import errno
 import gzip
@@ -25,7 +26,7 @@ class SpawnProcess(SubProcess):
 		"path_lookup", "pre_exec")
 
 	__slots__ = ("args",) + \
-		_spawn_kwarg_names
+		_spawn_kwarg_names + ("_log_file_real", "_selinux_type",)
 
 	_file_names = ("log", "process", "stdout")
 	_files_dict = slot_dict_class(_file_names, prefix="")
@@ -38,16 +39,6 @@ class SpawnProcess(SubProcess):
 		if self.fd_pipes is None:
 			self.fd_pipes = {}
 		fd_pipes = self.fd_pipes
-		fd_pipes.setdefault(0, sys.stdin.fileno())
-		fd_pipes.setdefault(1, sys.stdout.fileno())
-		fd_pipes.setdefault(2, sys.stderr.fileno())
-
-		# flush any pending output
-		for fd in fd_pipes.values():
-			if fd == sys.stdout.fileno():
-				sys.stdout.flush()
-			if fd == sys.stderr.fileno():
-				sys.stderr.flush()
 
 		self._files = self._files_dict()
 		files = self._files
@@ -61,28 +52,42 @@ class SpawnProcess(SubProcess):
 			logfile = self.logfile
 
 		null_input = None
-		fd_pipes_orig = fd_pipes.copy()
-		if self.background:
+		if not self.background or 0 in fd_pipes:
+			# Subclasses such as AbstractEbuildProcess may have already passed
+			# in a null file descriptor in fd_pipes, so use that when given.
+			pass
+		else:
 			# TODO: Use job control functions like tcsetpgrp() to control
 			# access to stdin. Until then, use /dev/null so that any
 			# attempts to read from stdin will immediately return EOF
 			# instead of blocking indefinitely.
 			null_input = open('/dev/null', 'rb')
 			fd_pipes[0] = null_input.fileno()
-		else:
-			fd_pipes[0] = fd_pipes_orig[0]
+
+		fd_pipes.setdefault(0, sys.stdin.fileno())
+		fd_pipes.setdefault(1, sys.stdout.fileno())
+		fd_pipes.setdefault(2, sys.stderr.fileno())
+
+		# flush any pending output
+		for fd in fd_pipes.values():
+			if fd == sys.stdout.fileno():
+				sys.stdout.flush()
+			if fd == sys.stderr.fileno():
+				sys.stderr.flush()
 
 		# WARNING: It is very important to use unbuffered mode here,
 		# in order to avoid issue 5380 with python3.
 		files.process = os.fdopen(master_fd, 'rb', 0)
 		if logfile is not None:
 
+			fd_pipes_orig = fd_pipes.copy()
 			fd_pipes[1] = slave_fd
 			fd_pipes[2] = slave_fd
 
 			files.log = open(_unicode_encode(logfile,
 				encoding=_encodings['fs'], errors='strict'), mode='ab')
 			if logfile.endswith('.gz'):
+				self._log_file_real = files.log
 				files.log = gzip.GzipFile(filename='', mode='ab',
 					fileobj=files.log)
 
@@ -146,7 +151,16 @@ class SpawnProcess(SubProcess):
 		return os.pipe()
 
 	def _spawn(self, args, **kwargs):
-		return portage.process.spawn(args, **kwargs)
+		spawn_func = portage.process.spawn
+
+		if self._selinux_type is not None:
+			spawn_func = portage.selinux.spawn_wrapper(spawn_func,
+				self._selinux_type)
+			# bash is an allowed entrypoint, while most binaries are not
+			if args[0] != BASH_BINARY:
+				args = [BASH_BINARY, "-c", "exec \"$@\"", args[0]] + args
+
+		return spawn_func(args, **kwargs)
 
 	def _output_handler(self, fd, event):
 
@@ -196,7 +210,12 @@ class SpawnProcess(SubProcess):
 					buf.tofile(files.log)
 				except TypeError:
 					# array.tofile() doesn't work with GzipFile
-					files.log.write(buf.tostring())
+					try:
+						# Python >=3.2
+						data = buf.tobytes()
+					except AttributeError:
+						data = buf.tostring()
+					files.log.write(data)
 				files.log.flush()
 			else:
 				self._unregister()
@@ -223,3 +242,9 @@ class SpawnProcess(SubProcess):
 
 		self._unregister_if_appropriate(event)
 
+	def _unregister(self):
+		super(SpawnProcess, self)._unregister()
+		if self._log_file_real is not None:
+			# Avoid "ResourceWarning: unclosed file" since python 3.2.
+			self._log_file_real.close()
+			self._log_file_real = None
